/*
 * wmdimp
 * Copyright (c) 2003 Daniele Tamino <aslo@wmdimp.cjb.net>
 * 
 * Based on wmmp3: Copyright (c)1999 Patrick Crosby <xb@dotfiles.com>.
 *
 * This software covered by the GPL.  See COPYING file for details.
 *
 * main.c
 *
 * This is the main body of the application.  Handles all initialization
 * plus the X event loop.
 *
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pwd.h>
#include <sys/types.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <pthread.h>
#include <assert.h>
#include <ctype.h>

#include "vect.h"
#include "buttons.h"
#include "wmgeneral.h"
#include "player.h"
#include "parsecfg.h"
#include "timer.h"
#ifdef HAVE_LIBMAD
#include "output.h"
#endif
#include "debug.h"

#define streq(s1, s2) \
        (strcmp(s1, s2) == 0)

#include "main.h"

volatile int terminate = 0;

opt_hash* loadconfig();
void show_help();
void show_version();
int check_options(int argc, char *argv[]);
int handle_button_press(int x, int y, int mousebtn);
void handle_button_release(int i, int x, int y, int mousebtn);
void wake_up_main_thread(long param);

/* a thread to handle signals without having to deal only
   with sig_atomic_t variables */
void* signal_thread(void* param)
{
    sigset_t *set = (sigset_t*) param;
    int sig;
    while(!terminate) {
        sigwait(set, &sig);
        switch(sig) {
        case SIGINT:
        case SIGTERM:
            DBGMSG("Caught signal %s: exiting\n",
                   sig==SIGTERM ? "SIGTERM" : "SIGINT");
            terminate = 1;
            wake_up_main_thread(WUMT_NOTHING);
            break;
#if /* !defined(NDEBUG) && */ DBG_MEMORY
        case SIGUSR1: {
            static int mtrace_enab = 0;
            if(mtrace_enab) {
                DBGMSG("Disabling malloc tracing\n");
                muntrace();
                mtrace_enab = 0;
            } else {
                DBGMSG("Enabling malloc tracing\n");
                mtrace();
                mtrace_enab = 1;
            }
            break;
        }
#endif
#ifndef NDEBUG
        case SIGHUP: {
            /* generate a SIGSEGV to test the debugging facility */
            int *addr = (int*) -1;
            *addr = 0; /* from gdb it's possible to change
                          addr and let the program continue  */
            break;
        }
#endif
            /* other signals to handle? */
        }
    }
    return NULL;
}

static pthread_t sig_thr = (pthread_t) 0;
void signal_init(void)
{
    static sigset_t signals;
    int err;
    sigemptyset(&signals);
    sigaddset(&signals, SIGTERM);
    sigaddset(&signals, SIGINT);
#if /* !defined(NDEBUG) && */ DBG_MEMORY
    sigaddset(&signals, SIGUSR1);
#endif
    DBG(sigaddset(&signals, SIGHUP));
#if 0/*ndef NDEBUG*/
    sigaddset(&signals, SIGSEGV); /* so we can inspect it with gdb */
#endif
   
    sigprocmask(SIG_BLOCK, &signals, NULL);
    err = pthread_create(&sig_thr, NULL, signal_thread, &signals);
    if(err != 0) {
        fprintf(stderr, "signal_init (pthread_create): %s\n", strerror(err));
        exit(err);
    }
}
void signal_finish(void)
{
    if(sig_thr != (pthread_t) 0) {
        pthread_cancel(sig_thr);
        pthread_join(sig_thr, NULL);
    }
}

/* micro widget "library" */

typedef void (*handle_click_f)(int wgt, int btn, int x, int y, void* data);
typedef struct widget
{
    handle_click_f hnd_press;
    handle_click_f hnd_release;
    int disabled; /* 1=do not send events */
    void* data;
} widget;

widget wgt_tab[];

void widget_disable(int wgt)
{
    button_gray(wgt);
    button_redraw(wgt);
    wgt_tab[wgt].disabled = 1;
}

void widget_enable(int wgt)
{
    button_up(wgt);
    button_redraw(wgt);
    wgt_tab[wgt].disabled = 0;
}

/* normal push button */

typedef void (*action_f)(int wgt);
typedef struct pushbtn
{
    action_f act, alt; /* action (left click) and alternative action
                        (right click) */
    int pressed;
} pushbtn;

void pushbtn_press(int wgt, int btn,int x, int y, void* data)
{
    pushbtn* pb = (pushbtn*) data;
    if(!pb->pressed) {
        button_down(wgt);
        button_redraw(wgt);
        pb->pressed = 1;
    }
    (void)btn; (void) x; (void) y; /* not used */
}

void pushbtn_release(int wgt, int btn, int x, int y, void* data)
{
    pushbtn* pb = (pushbtn*) data;
    if(pb->pressed) {
        int relx, rely;
        button_up(wgt);
        button_redraw(wgt);
        pb->pressed = 0;
        relx = x - btn_pos[wgt].x;
        rely = y - btn_pos[wgt].y;
        /* launch action only if the release was in the button */
        if(relx >= 0 && relx < btn_pos[wgt].w &&
           rely >= 0 && rely < btn_pos[wgt].h) {
            if(btn != Button1 && pb->alt) pb->alt(wgt);
            else pb->act(wgt);
        }
    }
}

/* toggle button */

typedef void (*toggle_f)(int wgt, int value);
typedef struct togglebtn
{
    toggle_f tog;
    int pressed;
} togglebtn;

void togglebtn_press(int wgt, int btn, int x, int y, void* data)
{
    togglebtn* tb = (togglebtn*) data;
    tb->pressed = !tb->pressed;
    if(tb->pressed) button_down(wgt);
    else button_up(wgt);
    button_redraw(wgt);
    tb->tog(wgt, tb->pressed);
    (void) btn; (void) x; (void) y;
}

/* widgets actions */

void handle_stop(int wgt)
{
    (void) wgt;
    button_up(B_PLAY);  /* raise play */
    button_redraw(B_PLAY);
    widget_enable(B_CHDIR);
    widget_disable(B_PAUSE);
    user_stop();
}

void handle_terminate(int wgt)
{
    handle_stop(wgt);
    terminate = 1;
}

void handle_back(int wgt)
{
    (void) wgt;
    if(is_paused()) {
        button_up(B_PAUSE);
        button_down(B_PLAY);
        button_redraw(B_PAUSE);
        button_redraw(B_PLAY);
    }
    user_back();
    if(is_playing()) show_list_number(500000); /* half a second */
    else show_list_number(3000000); /* 3 sec */
}

void handle_next(int wgt)
{
    (void) wgt;
    if(is_paused()) {
        button_up(B_PAUSE);
        button_down(B_PLAY);
        button_redraw(B_PAUSE);
        button_redraw(B_PLAY);
    }
    user_next();
    if(is_playing()) {
        DBGMSG("already playing, showing list num for 0.5 sec\n");
        show_list_number(500000); /* half a second */
    }
    else {
        DBGMSG("not playing, showing list num for 3.0 sec\n");
        show_list_number(3000000); /* 3 sec */
    }
}

void handle_chdir_up(int wgt) { (void)wgt; list_up(); }
void handle_chdir_down(int wgt) { (void)wgt; list_down(); }

void handle_random(int wgt, int value) {
    (void) wgt;
    if(value) random_on();
    else random_off();
}
void handle_repeat(int wgt, int value) {
    (void) wgt;
    if(value) repeat_on();
    else repeat_off();
}

void handle_pause(int wgt, int btn, int x, int y, void* data)
{
    (void) btn; (void) x; (void) y; (void) data;
    if(is_playing() || is_paused()) {
        button_down(wgt);
        user_pause();
        if(is_playing())
            button_down(B_PLAY);
        else 
            button_up(B_PLAY);
        button_redraw(B_PLAY);
        button_redraw(wgt);
    }
}

void handle_pause_release(int wgt, int btn, int x, int y, void* data)
{
    (void) btn; (void) x; (void) y; (void) data;
    if(is_playing()) {
        button_up(wgt);
        button_redraw(wgt);
    }
}

void handle_play(int wgt, int btn, int x, int y, void* data)
{
    if(!is_playing()) {
        button_down(wgt);
        widget_enable(B_PAUSE);
        widget_disable(B_CHDIR);
        if(!user_play()) {
            DBGMSG("handle_play: user_play() failed.\n");
            widget_disable(B_PAUSE);
            button_up(wgt);
            widget_enable(B_CHDIR);
        }
        button_redraw(wgt);
    } else
        handle_pause(wgt, btn, x, y, data);
}

void handle_title(int wgt, int btn, int x, int y, void* data)
{
    (void) wgt; (void) btn; (void) x; (void) y; (void) data;
    if(message_showing(msg_titlebar))
        unshow_message(msg_titlebar);
    else
        toggle_scroll();
}

void handle_listname()
{
    if(message_showing(msg_listbar)) {
        unshow_message(msg_listbar);
        if(is_playing())
            show_time();
    } else {
        /* show list name for 3 seconds */
        show_message(msg_listbar, 3000000, curr_list_name());
    }
}

void handle_listbar(int wgt, int btn, int x, int y, void* data)
{
    (void) wgt; (void) x; (void) y; (void) data;
    if(btn == Button1) time_remain_toggle();
    else handle_listname();
}

void handle_progbar(int wgt, int btn, int x, int y, void* data)
{
    (void) btn; (void) y; (void) data;
    /* jump to clicked position */
    if(!is_playing()) {
        button_down(B_PLAY);
        button_redraw(B_PLAY);
        widget_enable(B_PAUSE);
        widget_disable(B_CHDIR);
    }
    user_jump(x - btn_pos[wgt].x, btn_pos[wgt].w);
}

pushbtn btn_stop = { handle_stop, handle_terminate, 0 };
pushbtn btn_back = { handle_back, NULL, 0 };
pushbtn btn_next = { handle_next, NULL, 0 };
pushbtn btn_chdir = { handle_chdir_down, handle_chdir_up, 0 };
togglebtn btn_rand = { handle_random, 0 };
togglebtn btn_repeat = { handle_repeat, 0 };
widget wgt_tab[] = {
    /* { press, release, disabled, data }  */
    /* B_STOP */     { pushbtn_press, pushbtn_release, 0, &btn_stop},
    /* B_PLAY  */    { handle_play, NULL, 0, NULL},
    /* B_PAUSE */    { handle_pause, handle_pause_release, 0, NULL },
    /* B_BACK */     { pushbtn_press, pushbtn_release, 0, &btn_back},
    /* B_NEXT */     { pushbtn_press, pushbtn_release, 0, &btn_next},
    /* B_CHDIR */    { pushbtn_press, pushbtn_release, 0, &btn_chdir},
    /* B_RAND */     { togglebtn_press, NULL, 0, &btn_rand },
    /* B_REPEAT */   { togglebtn_press, NULL, 0, &btn_repeat },
    /* B_TITLE */    { handle_title, NULL, 0, NULL },
    /* B_LISTNAME */ { handle_listbar, NULL, 0, NULL },
    /* B_PROGBAR */  { handle_progbar, NULL, 0, NULL },
    /* terminator */ { NULL, NULL, 0, NULL }
};

int handle_button_press(int x, int y, int mousebtn)
{
    int i = CheckMouseRegion(x, y);

    if(!wgt_tab[i].disabled) {
        if(i >= 0 && i < B_NUM && wgt_tab[i].hnd_press)
            wgt_tab[i].hnd_press(i, mousebtn, x, y, wgt_tab[i].data);
        DBG(else DBGMSG("unknown button pressed\n"));
    }

    RedrawWindow();
    return(i);
}

void handle_button_release(int i, int x, int y, int mousebtn)
{
    if(!wgt_tab[i].disabled) {
        if(i >= 0 && i < B_NUM && wgt_tab[i].hnd_release)
            wgt_tab[i].hnd_release(i, mousebtn, x, y, wgt_tab[i].data);
        DBG(else DBGMSG("unknown button released\n"));
    }
}

opt_hash* loadconfig(void)
{
    struct vect specs;
    opt_hash* ht;
    player_factory** pl;
    int err;

    if(!vect_init(&specs, sizeof(type_desc_spec*)))
        return NULL;
    /* predefined type descriptions */
    if(!vect_incr(&specs, 1)) return NULL;
    ((const type_desc_spec**)specs.v)[0] = &player_specs;
#ifdef HAVE_LIBMAD /* output module is *currently* used only by player_mad.c */
    if(!vect_incr(&specs, 1)) return NULL;
    ((const type_desc_spec**)specs.v)[1] = &output_specs;
#endif
    /* each specific player's own spec */
    for(pl = player_tab; *pl != NULL; pl++) {
        if(!vect_incr(&specs, 1)) return NULL;
        *_vect_get_last(&specs, const type_desc_spec*) = &(*pl)->spec;
    }
    /* terminator */
    if(!vect_incr(&specs, 1)) return NULL;
    *_vect_get_last(&specs, type_desc_spec*) = NULL;
    
    ht = malloc(sizeof(*ht));
    if(!ht) return NULL;
    opt_hash_init(ht);
    err = load_config_file_multi("wmdimprc", (const type_desc_spec**)
                                 specs.v, ht, 0);
    free(specs.v);

    if(err != CFGERR_NONE) {
        fprintf(stderr, "Error loading config file: %s\n", get_cfg_err(err));
        if(err == CFGERR_SPECS) {
            fprintf(stderr, "BUG: Internal error, aborting :-(((((\n");
            abort();
        }
        if(err == CFGERR_SYSTEM && errno == ENOMEM)
            return NULL;
    }

    return ht;
}


const char version_str[] =
    "wmdimp -- Window Maker DIgital Music Player -- ver. "
        PACKAGE_VERSION "\n"
    "Copyright (C) 2003, Daniele Tamino <" PACKAGE_BUGREPORT ">\n"
    "\tBased on wmmp3 ver. 0.12 by Patrick Crosby\n";
const char freesoftware_str[] = 
    "This is free software, and you are welcome to redistribute it\n"
    "under certain conditions; also, it comes with ABSOLUTELY NO WARRANTY.\n"
    "See the GNU General Public License for more details (file COPYING).\n";

void show_help()
{
    puts(version_str);
    printf("At this moment, there are no command line options that do\n");
    printf("anything special:\n");
    printf("\t-h,--help:  print this message\n");
    printf("\t-v,--version:  print version info\n");
    printf("All options are set in ~/.wmdimprc.  See sample.wmdimprc in the\n");
    printf("distribution for more info.\n");
}

void show_version()
{
    puts(version_str);
    puts(freesoftware_str);
}

int check_options(int argc, char *argv[])
{
    int option_entered = 0;
    int i;
    
    for (i = 1; i < argc; i++) {
        if (streq(argv[i], "-h")) {
            option_entered = 1;
            show_help();
        }
        else if (streq(argv[i], "--help")) {
            option_entered = 1;
            show_help();
        }
        else if (streq(argv[i], "-v") || streq(argv[i], "--version")) {
            option_entered = 1;
            show_version();
        }
    }

    return option_entered;
}

Atom wm_delete_window;
long wm_wakeup;

void wake_up_main_thread(long param)
{
    XClientMessageEvent wakeup_event;
    wakeup_event.type = ClientMessage;
    wakeup_event.display = display;
    wakeup_event.window = win;
    wakeup_event.format = 32;
    wakeup_event.data.l[0] = wm_wakeup; 
    wakeup_event.data.l[1] = param;
  
    XLockDisplay(display); /* synchronize threads */
    XSendEvent(display, win, False, NoEventMask, (XEvent*)&wakeup_event);
    XFlush(display);
    XUnlockDisplay(display); /* synchronize threads */
}

int main(int argc, char *argv[])
{
    char mask_bits[64 * 64];
    int mask_width = 64;
    int mask_height = 64;
    XEvent Event;
    int i;
    int btn_num = 0;
    opt_hash* ht;

    DBG(dbg_init(argc, argv));

    if (check_options(argc, argv)) {
        exit(1);
    }

    ht = loadconfig();
    if(!ht) {
        fprintf(stderr, "Error loading config file: %s\n", strerror(ENOMEM));
        return -1;
    }

    /* set up the display area for wmdimp */
    createXBMfromXPM(mask_bits, wmdimp_pixmap, mask_width, mask_height);
    openXwindow(argc, argv, wmdimp_pixmap, mask_bits, mask_width, mask_height);

    font_init();
    draw_string("wmdimp", btn_pos[B_DIR].x, btn_pos[B_DIR].y, btn_pos[B_DIR].w);

    i = 0;
    for (i = 0; i < B_NUM; i++)
        AddMouseRegion(i, btn_pos[i].x, btn_pos[i].y,
                       btn_pos[i].x + btn_pos[i].w,
                       btn_pos[i].y + btn_pos[i].h);
    button_gray(B_PAUSE); /* pause is initially disabled */

    wm_delete_window = XInternAtom (display, "WM_DELETE_WINDOW", False);
    XSetWMProtocols(display, win, &wm_delete_window, 1);
    /* i could use every long integer, but this is cooler! */
    wm_wakeup = (long)XInternAtom(display, "wake up", False);

    player_init(ht);

    /* this should free the hash and every tree (and everything else).
       do we need it anymore? */
    opt_hash_free(ht);
    free(ht);
    ht = NULL;

    signal_init();
   /* event loop */
    while (!terminate) {
        XNextEvent(display, &Event);
        if(Event.xany.window != win && Event.xany.window != iconwin)
            continue;
        switch (Event.type) {
        case Expose: {
            XExposeEvent* e=&Event.xexpose;
            RedrawRegion(e->x, e->y, e->width, e->height);
            break;
        }
        case ClientMessage:
            if((unsigned)Event.xclient.data.l[0] == wm_delete_window) {
                terminate = 1;
            } else if(Event.xclient.data.l[0] == wm_wakeup) {
                /* my own event to wake up the main thread */
                switch(Event.xclient.data.l[1]) {
                case WUMT_NOTHING: /* wake up "for no reason" (to check terminate) */
                    break;
                /* no other messages used by now  */
                }
            }
            break;
        case DestroyNotify:
            player_finish();
            XCloseDisplay(display);
            return 0;
        case ButtonPress:
            btn_num = handle_button_press(Event.xbutton.x, Event.xbutton.y,
                                          Event.xbutton.button);
            break;
        case ButtonRelease:
            handle_button_release(btn_num, Event.xbutton.x, Event.xbutton.y,
                                  Event.xbutton.button);
            break;
        default:
            break;
        }
    }
    DBGMSG("main: stopping signal thread...\n");
    signal_finish();
    DBGMSG("main: destroying player...\n");
    player_finish();
    DBGMSG("main: destroying windows..\n");
    XDestroyWindow(display, iconwin);
    XDestroyWindow(display, win);
    XCloseDisplay(display);
    DBGMSG("main: bye! see you soon!\n");
    return 0;
}
