/*
 * wmdimp
 * Copyright (c) 2003 Daniele Tamino <aslo@wmdimp.cjb.net>
 *
 * This software covered by the GPL.  See COPYING file for details.
 *
 * player.h
 *
 * Interface for a generic player and functions to play songs (header file)
 *
 */
#ifndef playerH
#define playerH

#include "parsecfg.h"
#include <pthread.h>

extern const type_desc_spec player_specs; /* player config option */

typedef enum play_states {
  st_stop=0,
  st_pause=1,
  st_play=2
} play_states;

/* used to store current position and total length (last pos + 1) */
typedef struct song_pos {
  long frames;
  float seconds;
} song_pos;

/*
  N.B. In this context, a "frame" is, by definition, the smallest unit a
  specific player can seek into a specific song (==> this unit depends on
  the player and the song being played).
  If you want to know/seek a position with the highes accuracy, use frames,
  else you're better use time (seconds).
  Talking about mp3 and similar files, a frame will usually corresponds to
  the "frame" defined by the compression format.
*/

enum player_callback_values {
  CB_UPDATE,     /* elapsed time/frames update */
  CB_SONGCHANGE, /* song change (automatic or after calling play()) */
  CB_STATUS,     /* status change (automatic or after calling a method) */
  CB_TITCHANGE,  /* song title change (arg is the title, a const char*) */
  CB_ERROR,  /* an error occurred (arg is a string that describes it) */
};

struct player;

/* N.B. callback is NOT allowed to call player methods's that changes
   the play status (setup, play, pause, stop) or position (jump*),
   nor it should block (otherwise, everything might happen, especially
   deadlocks).
   When what == CB_SONGCHANGE, it may return 1 if there's a new song to
   be played soon by the main thread, or 0 if the player should stop.
*/
typedef int (*player_cb_func)(struct player* pl, int what,
                   void* param, void* arg);

typedef struct player
{
  /* fields are to be considered read-only, except for the methods below */
  const char* name; /* name of the player (mpg123, ogg player, ???) */
  volatile int status; /* current status (one of play_states) */

  /* methods - they return 0 on failure and != 0 otherwise */

  void (*finish)(struct player*); /* destructor */
 
  /* test if the file is in the correct format for this player */
  int (*test)(struct player*, const char* fn);
  /* setup `filename' for playing and, if current status is not st_stop,
     or 'play' is != 0, start playing it.
     Stores in 'title' (if != NULL) a malloc'ed buffer with the title to show
     (may be something like 'title - album - artist', not just the song title).
     It's your responsibility to free the title buffer (if requested).
     Use me->setup(me, NULL, NULL); to release everything.
  */
  int (*setup)(struct player*, const char* fn, char ** title, int play);
  int (*play)(struct player*);  /* play asynchronously */
  /* play synchronously (= this call returns when playing ends) */
  /* N.B. synchronous playing is done especially to enable profiling */
  int (*play_sync)(struct player*);
  /* 'pause' will pause if playing and resume if already paused
     (and do nothing if stopped) */
  int (*pause)(struct player*);
  int (*stop)(struct player*);  /* guess ;) */

  /* jumps 'f' frames from the beginning */
  int (*jump_frames)(struct player*, long f);
  /* jumps 'sec' seconds from the beginning */
  int (*jump_time)(struct player*, long sec);
  /* jumps to `position/total' of the total file (useful for progress bar) */
  int (*jump)(struct player*, long position, long total);
  /* N.B. jump* methods will start playing if status != st_play */

  /* query total length (total number of frames/seconds) */
  int (*get_length)(struct player*, song_pos* len);
  /* query current position */
  int (*get_pos)(struct player*, song_pos* pos);

  /* set callback */
  void (*set_callback)(struct player*, player_cb_func cb, void* param);
} player;

typedef struct player_factory {
    /* initialize static structures (loads vaules from config) */
    int (*static_init)(opt_hash* ht);
    /* releases static structures */
    void (*static_finish)(void);
    /* malloc's and init's a new player of the correct type */
    player* (*create)(void);
    /* specific player's config options */
    type_desc_spec spec;
} player_factory;

static inline void player_destroy(player* pl) {
    if(pl) {
        pl->finish(pl);
        free(pl);
    }
}

/* NULL terminated list of player factories (in playertab.c) */
extern player_factory* player_tab[];
/* number of elements in player_tab (without the terminator) */
extern const int player_tab_sz;

static inline int static_init_all_players(opt_hash* ht) {
    player_factory** pf;
    int ret = 1;
    for(pf = player_tab; *pf != NULL; pf++)
    ret = ret && (*pf)->static_init(ht);
    return ret;
}
static inline void static_finish_all_players(void) {
    player_factory** pf;
    for(pf = player_tab; *pf != NULL; pf++)
    (*pf)->static_finish();
}

extern player* curr_player;
extern pthread_mutex_t player_mtx;

void player_init(opt_hash* ht);   /* was init_ctl */
void player_finish(void); /* was finish_ctl */

/* shortcuts to get current player state */
static inline int is_playing(void) {
  int ret;
  pthread_mutex_lock(&player_mtx);
  ret = curr_player != NULL && curr_player->status == st_play;
  pthread_mutex_unlock(&player_mtx);
  return ret;
}
static inline int is_paused(void) {
  int ret;
  pthread_mutex_lock(&player_mtx);
  ret = curr_player != NULL && curr_player->status == st_pause;
  pthread_mutex_unlock(&player_mtx);
  return ret;
}
static inline int is_stopped(void) {
  int ret;
  pthread_mutex_lock(&player_mtx);
  ret = curr_player != NULL && curr_player->status == st_stop;
  pthread_mutex_unlock(&player_mtx);
  return ret;
}

extern pthread_mutex_t callback_mtx;
/* to be called when the user requests that action */
static inline int user_pause() { /* was 'play_pause' */
  int ret;
  pthread_mutex_lock(&callback_mtx);
  pthread_mutex_lock(&player_mtx);
  ret = (curr_player && curr_player->pause(curr_player));
  pthread_mutex_unlock(&player_mtx);
  pthread_mutex_unlock(&callback_mtx);
  return ret;
}
int user_play(void);
int user_stop(void);  /* was 'stop' */
int user_next(void);  /* was 'next' */
int user_back(void);  /* was 'back' */

#if 0
/* simple shortct functions */
static inline int user_jump(long progress, long total) { /* was 'jump' */
  if(!curr_player) return 0;
  return curr_player->jump(curr_player, progress, total);
}
static inline int user_jump_frame(long frame) {
  if(!curr_player) return 0;
  return curr_player->jump_frames(curr_player, frame);
}

static inline int user_jump_time(long sec) {
  if(!curr_player) return 0;
  return curr_player->jump_time(curr_player, sec);
}
#else
int user_jump(long progress, long total); /* was 'jump' */
int user_jump_frame(long frame);
int user_jump_time(long sec);
#endif

extern volatile int show_msg_tit; /* >1 = showing a message in the titlebar */
extern volatile int show_msg_lst; /* >1 = showing a message in the listbar */

/* config functions */
void set_file_ext(opt_tree_node* array); /* was 'set_mp3ext' */
void set_list_ext(opt_tree_node* array); /* was 'set_playlistext' */
void set_alwaysscroll(int as);
void set_scrolldelay(float d);
void set_shownumdelay(float d); /* when browsing a list, how long to show number */
void set_lists(opt_tree_node* array);
void set_sorting(char *s); /* unsorted (or false), sorted (or true), reversed */

/* list functions */
void list_up(/* int button_num */); /* was 'dir_up' */
void list_down(/* int button_num */); /* was 'dir_down' */
const char* curr_list_name(void);

/* other display/status functions */
void turn_on_scroll(void);
void toggle_scroll(void);
void show_title(void);
void time_remain_toggle(void);

void show_progress_bar(void);
void show_time(void);
void show_list_name(void);
void show_list_number(unsigned long usec);

#include <stdarg.h>

int wgt_printf(int widget, const char* fmt, ...)
#ifdef __GNUC__
     __attribute__ ((__format__ (__printf__, 2, 3)))
#endif
;
int wgt_vprintf(int widget, const char* fmt, va_list vl)
#ifdef __GNUC__
     __attribute__ ((__format__ (__printf__, 2, 0)))
#endif
;

enum msg_loc { msg_titlebar, msg_listbar };

/* show a message in the specified location (where) for "us" microseconds
   (unless another message arrives before) */
int show_message(enum msg_loc where, unsigned long us,
         const char* fmt, ...)
#ifdef __GNUC__
     __attribute__ ((__format__ (__printf__, 3, 4)))
#endif
;
int show_message_vl(enum msg_loc where, unsigned long us,
            const char* fmt, va_list vl)
#ifdef __GNUC__
     __attribute__ ((__format__ (__printf__, 3, 0)))
#endif
;

/* reset flag to let normal usage of the bar (show title/time).
   This doesn't mean it will show immediatedly. */
void unshow_message(enum msg_loc where);
/* return whether there's a message showing */
int message_showing(enum msg_loc where);

#if 0
void random_toggle(int button_num);
void repeat_toggle(int button_num);
#else
extern int random_flag, repeat_flag;
void random_on(void);
void random_off(void);
void repeat_on(void);
void repeat_off(void);
#endif

#endif
