/*
 * wmdimp
 * Copyright (c) 2003 Daniele Tamino <aslo@wmdimp.cjb.net>
 * This software covered by the GPL.  See COPYING file for details.
 *
 * Program to test (debug and profile) a player (maily player_mad.c)
 * in a more convenient way.
 *
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include <signal.h>
#include <time.h>
#include <sys/times.h>
#include <sys/stat.h>

#include "parsecfg.h"
#include "player.h"
#include "output.h"

void loadconfig(char* cfg, opt_hash *ht)
{
    const type_desc_spec** specs;
    int i, err;

    specs = (const type_desc_spec**)
        malloc((player_tab_sz + 2) * sizeof(*specs));
    if(!specs) exit(-1);
    /* predefined type descriptions (only one) */
    specs[0] = &output_specs;
    /* each specific player's own config specification */
    for(i = 0; i < player_tab_sz; i++)
        specs[i + 1] = &player_tab[i]->spec;
    specs[i+1] = NULL; /* terminator */
    
    opt_hash_init(ht);
    err = load_config_file_multi(cfg, specs, ht,
                                 CFL_SEARCHWD|CFL_IGNUNKNOWN);
    free(specs);

    if(err != CFGERR_NONE) {
        fprintf(stderr, "Error loading config file: %s\n", get_cfg_err(err));
        if(err == CFGERR_SPECS) {
            fprintf(stderr, "BUG: Internal error, aborting :-(((((\n");
            abort();
        }
        if(err == CFGERR_SYSTEM && errno == ENOMEM)
            exit(-1);
    }
}

static volatile sig_atomic_t terminate = 0;
void signal_term(int sig) {
    (void) sig; terminate = 1;
}

void my_callback(struct player* pl, int what, void* param, void* arg)
{
    static float last_secs = -1.0;
    float diff;
    int bufsz, bufspc, perc;
    song_pos pos, len;
    (void) param;
    switch(what) {
    case CB_UPDATE:
        pl->get_pos(pl, &pos);
        pl->get_length(pl, &len);
        diff = pos.seconds - last_secs;
        if(output_bufinfo(&bufsz, &bufspc))
            perc = 100 - bufspc * 100 / bufsz;
        else
            perc = 0;
        if(diff < 0.0 || diff >= 1.0) {
            int me, mr, se, sr;
            me = (int) pos.seconds / 60;
            se = (int) (pos.seconds - me * 60);
            mr = (int) (len.seconds - pos.seconds) / 60;
            sr = (int) ((len.seconds - pos.seconds) - mr * 60);
            printf("elapsed %02d:%02d (remaining %02d:%02d) buffer: %d%% "
                   "(%d/%d samples)\r",
                   me, se, mr, sr, perc, bufsz - bufspc, bufsz);
            fflush(stdout);
            last_secs = pos.seconds;
        }
        break;
    case CB_ERROR:
        fprintf(stderr, "%s: %s\n", pl->name, (char*)arg);
        break;
    }
    if(terminate)
        pl->stop(pl);
}

void show_used_time(struct tms* start, struct tms* end)
{
    printf("Time used (only) while playing:\n");
    printf("\t%.2fs user, %.2fs system, %.2fs childs\n",
           (double)(end->tms_utime - start->tms_utime) / CLK_TCK,
           (double)(end->tms_stime - start->tms_stime) / CLK_TCK,
           (double)(end->tms_cstime - start->tms_cstime) / CLK_TCK);
}

int main(int argc, char** argv)
{
    char* cfg = "testplayrc";
    char* fn = "test.mp3";
    char* tit = NULL;
    player_factory **pf;
    player* pl = NULL;
    opt_hash ht;
    struct stat st;
    int c;
    struct tms time, tend;
    
    while((c=getopt(argc, argv, "r:f:")) >= 0) {
        switch(c) {
        case 'r':
            cfg=optarg;
            break;
        case 'f':
            fn=optarg;
            break;
        }
    }

    if(stat(fn, &st) != 0) {
        perror(fn);
        return 1;
    }
    loadconfig(cfg, &ht);
    static_init_all_players(&ht);

    for(pf = player_tab; *pf != NULL; pf++) {
        if((pl = (*pf)->create()) && pl->test(pl, fn))
            break;
        if(pl) {
            pl->finish(pl);
            free(pl);
        }
    }
    if(pf == NULL) {
        fprintf(stderr, "%s: unsupported file format.\n", fn);
        return 2;
    }
    signal(SIGTERM, signal_term);
    signal(SIGINT, signal_term);
    pl->set_callback(pl, my_callback, NULL);
    printf("%s: setting up...\n", fn);
    pl->setup(pl, fn, &tit, 0);
    if(!tit) tit = fn;
    printf("%s: playing synchronously (if possible)\n", tit);

    /* do some benchmarking too */
    times(&time);
    pl->play_sync(pl);
    times(&tend);
    show_used_time(&time, &tend);

    printf("done!\n");
    pl->finish(pl);
    free(pl);
    return 0;
}
