/*
 * wmdimp
 * Copyright (c) 2003 Daniele Tamino <aslo@wmdimp.cjb.net>
 *
 * This software covered by the GPL.  See COPYING file for details.
 *
 * testtmr.c
 *
 * random test program to test timer.c
 *
 */
#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>
#include <stdarg.h>
#include <pthread.h>
#include <sys/time.h>

#include "timer.h"
#include "debug.h"

/* taken from GNU libc info guide */

/* Subtract the `struct timeval' values X and Y,
   storing the result in RESULT.
   Return 1 if the difference is negative, otherwise 0.  */
     
int timeval_subtract (struct timeval *result,
              struct timeval *x,
              struct timeval *y)
{
  /* Perform the carry for the later subtraction by updating Y. */
  if (x->tv_usec < y->tv_usec) {
    int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
    y->tv_usec -= 1000000 * nsec;
    y->tv_sec += nsec;
  }
  if (x->tv_usec - y->tv_usec > 1000000) {
    int nsec = (x->tv_usec - y->tv_usec) / 1000000;
    y->tv_usec += 1000000 * nsec;
    y->tv_sec -= nsec;
  }
     
  /* Compute the time remaining to wait.
     `tv_usec' is certainly positive. */
  result->tv_sec = x->tv_sec - y->tv_sec;
  result->tv_usec = x->tv_usec - y->tv_usec;
     
  /* Return 1 if result is negative. */
  return x->tv_sec < y->tv_sec;
}

pthread_mutex_t screen_mtx = PTHREAD_MUTEX_INITIALIZER;

#ifndef NDEBUG
/* declared in debug.c */
extern void print_time(FILE* out, struct timeval* t);
#else
void print_time(FILE* out, struct timeval* t)
{
    double sec = (t->tv_sec % 60) + (double)(t->tv_usec)/1e6;
    int min  = t->tv_sec / 60;
    int hour = min / 60;
    min %= 60;
    hour %= 24; /* we don't mind exact time, just intervals */
    if(hour > 0) fprintf(out, "%dh", hour);
    if(min > 0)  fprintf(out, "%dm", min);
    fprintf(out, "%.3fs", sec);
}
#endif

int msg_printf(char fill, const char* fmt, ...)
{
  struct timeval tv;
  int ret;
  va_list va;
  va_start(va, fmt);
  gettimeofday(&tv, NULL);
  printf("%c%c%c (", fill, fill, fill);
  print_time(stdout, &tv);
  printf("): ");
  ret = vprintf(fmt, va);
  va_end(va);
  return ret;
}

#define NUM_TIMERS 100
volatile int tmr_done = 0;
volatile int timer_ran[NUM_TIMERS]; /* 1 = ran, 0 = not ran */

#define EPSILON 0.02
/* check if the absolute value of diff is less than EPSILON */
int is_less_than_epsilon(struct timeval *diff)
{
  double x = diff->tv_sec + (double)diff->tv_usec / 1000000;
  if(x < 0) x = -x; /* get absolute value */
  return (x < EPSILON);
}

void my_timer(struct timeval* tv, void* param)
{
  struct timeval now, diff;
  int num = (int) param;
  int error;
  gettimeofday(&now, NULL);
  timeval_subtract(&diff, &now, tv);
  error = !is_less_than_epsilon(&diff);
  pthread_mutex_lock(&screen_mtx);
  if(error) fputs("\e[4;41m", stdout); /* bg red + hilight */
  
  msg_printf('*', "executing timer %d with ", num);
  print_time(stdout, &diff);
  fputs(" delay", stdout);
  if(error) fputs("\e[0m", stdout); /* reset colors */
  fputs("\n", stdout);
  fflush(stdout);
  pthread_mutex_unlock(&screen_mtx);
  tmr_done++; /* i hope this is atomic */
  timer_ran[num] = 1; /* this too */
}

/* generate a random number between 0 and max inclusive */
#ifndef NDEBUG
int randnum(int max) {
  int rnd = rand();
  int ret = (int) (((double)rnd / RAND_MAX) * max);
  assert(ret >= 0 && ret <= max);
  return ret;
}
#else
int randnum(int max) {
 return (int) (((double)rand() / RAND_MAX) * max);
}
#endif

#if 1
/* now testing timer_changetime too */
#define MAX_TIME (NUM_TIMERS/10) /* max 0.1 seconds per timer (avg) */
struct timer {
  struct timeval when;   /* when to execute it */
  struct timeval sched;  /* when to schedule it */
  int id;
} timer_tab[NUM_TIMERS];

int compare_timers(const void* a, const void* b)
{
  const struct timeval *ta = &(*(const struct timer**) a)->sched;
  const struct timeval *tb = &(*(const struct timer**) b)->sched;
  if(ta->tv_sec == tb->tv_sec) {
    if(ta->tv_usec == tb->tv_usec) return 0;
    else if(ta->tv_usec < tb->tv_usec) return -1;
    else return 1;
  } else if(ta->tv_sec < tb->tv_sec)
      return -1;
  else
    return 1;

}
void do_test(void)
{
  struct timeval now, diff;
  struct timespec ts;
  int i, j;
  struct timer* tmr_tab[NUM_TIMERS];
  
  /* fill address table */
  for(i = 0; i < NUM_TIMERS; i++)
    tmr_tab[i] = timer_tab + i;

  gettimeofday(&now, NULL);
  now.tv_sec += 1; /* add a second to do all precomputation */
  /* fill timer_tab */
  for(i = 0; i < NUM_TIMERS; i++) {
    int from_now = randnum(MAX_TIME);
    int rnd;
    timer_tab[i].when.tv_sec = now.tv_sec + from_now;
    timer_tab[i].when.tv_usec = randnum(999999);
    rnd = randnum(from_now);
    timer_tab[i].sched.tv_sec = now.tv_sec + rnd;
    if(rnd == from_now)
      timer_tab[i].sched.tv_usec = randnum(timer_tab[i].when.tv_usec);
    else
      timer_tab[i].sched.tv_usec = randnum(999999);
    if(rand() / (RAND_MAX/2) == 0) /* 1/2 times */
      timer_tab[i].id = -1;
    else
      /* this means to reschedule a random timer if there are some pending */
      timer_tab[i].id = 1;
  }
  /* sort it */
  qsort(tmr_tab, NUM_TIMERS, sizeof(tmr_tab[0]), compare_timers);

  /* show the generated table for debugging */
  for(i = 0; i < NUM_TIMERS; i++) {
    int error = timeval_subtract(&diff, &tmr_tab[i]->when, &tmr_tab[i]->sched);
    if(error) fputs("\e[4;41m", stderr); /* bg red + hilight */
    fprintf(stderr, "DBG: timer %d: sched=", i);
    print_time(stderr, &tmr_tab[i]->sched);
    fputs(", when=", stderr);
    print_time(stderr, &tmr_tab[i]->when);
    if(tmr_tab[i]->id > 0)
      fputs(" (reschedule)", stderr);
    if(error) fputs("\e[0m", stderr); /* reset colors */
    fputs("\n", stderr);
  }

  for(i = 0; i < NUM_TIMERS; i++) {
    gettimeofday(&now, NULL);
    if(timeval_subtract(&diff, &tmr_tab[i]->sched, &now) == 0) {
      ts.tv_sec = diff.tv_sec;
      ts.tv_nsec = diff.tv_usec * 1000;
      pthread_mutex_lock(&screen_mtx);
      msg_printf('-', "Sleeping for %.10g sec\n",
                 ts.tv_sec + 0.000000001*ts.tv_nsec);
      pthread_mutex_unlock(&screen_mtx);
      nanosleep(&ts, NULL);
    }
    if(tmr_tab[i]->id < 0 || i == 0) {
      pthread_mutex_lock(&screen_mtx);
      msg_printf('-', "Scheduling job %d for ", i);
      print_time(stdout, &tmr_tab[i]->when);
      tmr_tab[i]->id = timer_schedule(&tmr_tab[i]->when,
                                      my_timer, (void*) i);
      printf(" (job id %d)\n", tmr_tab[i]->id);
      pthread_mutex_unlock(&screen_mtx);
    } else {
      /* search a running timer (starting at a random location) */
      int start = randnum(i - 1);
      j = start;
      do {
        if(!timer_ran[j]) break;
        if(++j >= i) j = 0;
      } while(j != start);

      if(!timer_ran[j]) {
        /* Occasionally, the timer could run just now, so the following
           call is invalid: no problem, test timer robustness ;-) */
        pthread_mutex_lock(&screen_mtx);
        msg_printf('-', "Re-scheduling timer %d (job %d) for ",
                   j, tmr_tab[j]->id);
        print_time(stdout, &tmr_tab[i]->when);
        printf("\n");
        pthread_mutex_unlock(&screen_mtx);
        if(timer_changetime(tmr_tab[j]->id, &tmr_tab[i]->when) < 0) {
          pthread_mutex_lock(&screen_mtx);
          msg_printf('E', "rescheduling failed!\n");
          pthread_mutex_unlock(&screen_mtx);
        }
        timer_ran[i] = 1; /* mark this rescheduling as done */
        tmr_done++;
      } else {
        /* create a new one instead of rescheduling */
        pthread_mutex_lock(&screen_mtx);
        msg_printf('-', "(No running timers) Scheduling job %d for ", i);
        print_time(stdout, &tmr_tab[i]->when);
        tmr_tab[i]->id = timer_schedule(&tmr_tab[i]->when,
                                        my_timer, (void*) i);
        printf(" (job id %d)\n", tmr_tab[i]->id);
        pthread_mutex_unlock(&screen_mtx);
      }
    }
  }
  pthread_mutex_lock(&screen_mtx);
  msg_printf('-', "Waiting for timers to complete...\n");
  pthread_mutex_unlock(&screen_mtx);
  ts.tv_sec = 0;
  ts.tv_nsec = 100000000L;
  while(tmr_done < 20)
    nanosleep(&ts, NULL);
}
#else
/* this "random timer generator" is very random only in
 * the way it is coded :)
 * so I decided to write something better (the one above)
 */
int timer_id[NUM_TIMERS]; /* associate timer num with job ids */
void do_test(void)
{
  int i, j;
  struct timeval tv;
  struct timespec ts;
  long s, us;
  int timer_num = 0;
  
  while(timer_num < NUM_TIMERS) {
    gettimeofday(&tv, NULL);
    /* generate an event from now to 5 seconds after */
    s = rand() * 5 / RAND_MAX; /* low bits are't very random */
    us = rand() % 1000000;
    tv.tv_sec  += s;
    tv.tv_usec += us;
    if(tv.tv_usec >= 1000000) {
      tv.tv_usec -= 1000000;
      tv.tv_sec ++;
    }
    /* 1/4 times (avg) reschedule a timer if there is one pending */
    if(timer_num > tmr_done && rand() < RAND_MAX/4) {
      int x = rand() % timer_num;
      for(j = 0; j < timer_num; j++)
        if(!timer_ran[(j + x) % timer_num])
          break;
      j = (j + x) % timer_num;
      /* Occasionally, the timer could run just now, so the following
         call is invalid: no problem, test timer robustness ;-)
         N.B. There is always a pending timer to find.
      */
      msg_printf('-', "Re-scheduling timer %d (job %d) for now + %.7g sec\n",
                 j, timer_id[j], s + 0.000001*us);
      if(timer_changetime(timer_id[j], &tv) < 0)
        msg_printf('E', "rescheduling failed!\n");
    } else {
      int num = timer_num++;
      msg_printf('-', "Scheduling timer %d for now + %.7g sec\n",
                 num, s + 0.000001*us);
      timer_id[num] = timer_schedule(&tv, my_timer, (void*) (num));
    }
    /* wait a random time from 0 to half the time we genereted before */
    ts.tv_sec = rand() % ((s / 2) + 1);
    ts.tv_nsec = rand() % 1000000000;
    msg_printf('-', "Sleeping for %.10g sec\n", ts.tv_sec + 0.000000001*ts.tv_nsec);
    nanosleep(&ts, NULL); /* don't care about EINTR, it just adds some random */
  }
  ts.tv_sec = 0;
  ts.tv_nsec = 100000000L;
  while(tmr_done < 20)
    nanosleep(&ts, NULL);
}
#endif

int main(int argc, char **argv)
{
  unsigned seed = time(0);

  setvbuf(stdout, NULL, _IOLBF, BUFSIZ); /* set line buffered output */
  setvbuf(stderr, NULL, _IOLBF, BUFSIZ);

  if(argc > 1)
    seed = strtol(argv[1], NULL, 0);

  printf("random generator seed: %u\n", seed);
  srand(seed);
  timer_start();
  do_test();
  timer_stop();
  return 0;
}
