/*
 * wmdimp
 * Copyright (c) 2003 Daniele Tamino <aslo@wmdimp.cjb.net>
 *
 * This software covered by the GPL.  See COPYING file for details.
 *
 * timer.h
 *
 * function to asynchronously run a handler at a given time (header file)
 *
 */
#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <sys/time.h>
#include <assert.h>
#include <errno.h>

#include "vect.h"
#include "timer.h"
#include "utils.h"
#define DBG_MESSAGES 0 /* I don't need 'em anymore  */
#include "debug.h" /* expands to nil if #defined NDEBUG */

static pthread_t timer_thr;
static int running = 0;
static volatile int terminate = 0;

static pthread_cond_t tmr_wakeup = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t tmr_mtx = PTHREAD_MUTEX_INITIALIZER;

typedef struct job {
    struct timeval tv;
    timer_handler th;
    void* param;
    int id; /* job id (index into jobs vector) */
} job;

typedef job* pq_ent;
struct vect pqueue; /* zero initialized */
struct vect jobs; /* translate job id --> pqueue index */
int job_last = -1; /* last job id allocated */

/* returns -1 if a < b, 1 if a > b and 0 if a == b */
inline int cmp_ent(pq_ent a, pq_ent b)
{
    if(a->tv.tv_sec == b->tv.tv_sec) {
        if(a->tv.tv_usec == b->tv.tv_usec) return 0;
        else if(a->tv.tv_usec < b->tv.tv_usec) return -1;
        else return 1;
    } else if(a->tv.tv_sec < b->tv.tv_sec)
        return -1;
    else
        return 1;
}


/* Memo:
 * i (start from 0): child1 --> 2i + 1
 *                   child2 --> 2i + 2
 *                   parent --> (i - 1) / 2
 */

/* move down the element i (the i-th element of the array pq, start from 0) */
void heap_down(pq_ent* pq, int* jb, int i, int size)
{
    pq_ent val = pq[i];
    int j = i; /* j is the smaller child of i */
    assert(i >= 0 && i < pqueue.size);
    while(2*j + 1 < size) {
        j = 2*i + 1; /* first child of i */
        if(j + 1 < size && cmp_ent(pq[j], pq[j + 1]) > 0)
            j++; /* second child is < first, use it */
        if(cmp_ent(val, pq[j]) <= 0)
            break;
        pq[i] = pq[j];
        assert(pq[i]->id >= 0 && pq[i]->id < jobs.size);
        jb[pq[i]->id] = i; /* update job id */
        i = j;
    }
    pq[i] = val;
    assert(pq[i]->id >= 0 && pq[i]->id < jobs.size);
    jb[pq[i]->id] = i; /* update job id */
}

void heap_up(pq_ent* pq, int* jb, int i)
{
    pq_ent val = pq[i];
    int j; /* j is the parent of i */
    assert(i >= 0 && i < pqueue.size);
    while(i >= 1) {
        j = (i-1) / 2;
        if(cmp_ent(pq[j], val) <= 0)
            break;
        pq[i] = pq[j];
        assert(pq[i]->id >= 0 && pq[i]->id < jobs.size);
        jb[pq[i]->id] = i;
        i = j;
    }
    pq[i] = val;
    assert(pq[i]->id >= 0 && pq[i]->id < jobs.size);
    jb[pq[i]->id] = i;
}

int pqueue_push(pq_ent e)
{
    pq_ent* pq;
    int* jb;
    if(!vect_incr(&pqueue, 1))
        return 0; /* very low memory! */
    /* allocate a new job id */
    if(pqueue.size > jobs.size) {
        if(!vect_incr(&jobs, 1))
            return 0; /* very low memory! */
        e->id = job_last = jobs.size - 1;
        jb = (int*) jobs.v;
    } else {
        /* find the free location (== -1) in jobs (it MUST exist) */
        int i = job_last;
        jb = (int*) jobs.v;
        do {
            if(jb[i] == -1) break;
            if(++i >= jobs.size) i = 0;
        } while(i != job_last);
        assert(jb[i] == -1); /* the loop should ALWAYS find a free id */
        e->id = job_last = i;
    }

    pq = (pq_ent*)pqueue.v;
    pq[pqueue.size - 1] = e;
    jb[e->id] = pqueue.size - 1;
    if(pqueue.size > 1)
        heap_up(pq, jb, pqueue.size - 1);
    return 1;
}

pq_ent pqueue_pop(void)
{
    pq_ent* pq = (pq_ent*)pqueue.v;
    int* jb = (int*) jobs.v;
    pq_ent e = pq[0];
    jb[e->id] = -1; /* mark this job as deleted */
    e->id = -1; /* not strictly necessary */
    pq[0] = pq[--pqueue.size];
#ifndef NDEBUG /* offensive programming (opposite of defensive) */
    /* try to catch (force sigsegv) the use of deleted items */
    pq[pqueue.size] = NULL;
#endif
    if(pqueue.size > 0)
        heap_down(pq, jb, 0, pqueue.size);
    return e;
}

inline int pqueue_init(void)
{
    if(pqueue.v == NULL) {
        vect_init(&pqueue, sizeof(pq_ent));
        if(pqueue.v == NULL) return 0;
    }
    if(jobs.v == NULL) {
        vect_init(&jobs, sizeof(int));
        if(jobs.v == NULL) return 0;
    }
    return 1;
}

inline void pqueue_destroy(void)
{
    if(pqueue.v != NULL) {
        free(pqueue.v);
        pqueue.v = NULL;
    }
    if(pqueue.v != NULL) {
        free(jobs.v);
        jobs.v = NULL;
    }
}

inline pq_ent pqueue_top(void)
{
    return *(pq_ent*) pqueue.v;
}

inline int pqueue_empty(void)
{
    return pqueue.size == 0;
}

/* taken from GNU libc info guide */

/* Subtract the `struct timeval' values X and Y,
   storing the result in RESULT.
   Return 1 if the difference is negative, otherwise 0.  */
     
static int timeval_subtract (struct timeval *result,
                 struct timeval *x,
                 struct timeval *y)
{
    /* Perform the carry for the later subtraction by updating Y. */
    if (x->tv_usec < y->tv_usec) {
        int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
        y->tv_usec -= 1000000 * nsec;
        y->tv_sec += nsec;
    }
    if (x->tv_usec - y->tv_usec > 1000000) {
        int nsec = (x->tv_usec - y->tv_usec) / 1000000;
        y->tv_usec += 1000000 * nsec;
        y->tv_sec -= nsec;
    }
     
    /* Compute the time remaining to wait.
     `tv_usec' is certainly positive. */
    result->tv_sec = x->tv_sec - y->tv_sec;
    result->tv_usec = x->tv_usec - y->tv_usec;
     
  /* Return 1 if result is negative. */
    return x->tv_sec < y->tv_sec;
}

int calc_time_to_wait(job* jb, struct timespec* twait)
{
    struct timeval now, then, diff;
    int neg;
    gettimeofday(&now, NULL);
    then = jb->tv;
    neg = timeval_subtract(&diff, &then, &now);
    twait->tv_sec = diff.tv_sec;
    twait->tv_nsec = diff.tv_usec * 1000;
    return neg;
}

static void* timer_thread(void* param)
{
    struct timespec tnext;
    job *jb = NULL;
    (void)param;
    pthread_mutex_lock(&tmr_mtx);

    DBGMSG("timer_thread started\n");
    while(!terminate) {
        if(pqueue_empty()) {
            DBGMSG("waiting for events (queue empty)...\n");
            pthread_cond_wait(&tmr_wakeup, &tmr_mtx);
        } else {
            int err;
            jb = pqueue_top(); /* O(1) */
            DBGMSG("waiting %.4gs for job %d...\n", __diff_from_now(&jb->tv), jb->id);
            tnext.tv_sec = jb->tv.tv_sec;
            tnext.tv_nsec = jb->tv.tv_usec * 1000;
            /* This call is not the most accurate way of waiting...
             * on linux it has only 20ms of accuracy.
             * Maybe it could be better to wait 20 msec less and then use
             * nanosleep to wait for the rest (although this could
             * block timer_schedule for 0-20ms sometimes)
             */
            err = pthread_cond_timedwait(&tmr_wakeup, &tmr_mtx, &tnext);
            if(err == ETIMEDOUT) {
                DBG(
                if(abs(__diff_from_now(&jb->tv)) > 0.02) {
                    DBGMSG("running job %d %.4gs late\n",
                           jb->id, __diff_from_now(&jb->tv));
                })
                jb->th(&jb->tv, jb->param);
                free(pqueue_pop()); /* O(log n) AFTER running the job */
            } else {
                DBGMSG("woken up by someone...\n");
            }
        }
    }
    pthread_mutex_unlock(&tmr_mtx);
    DBGMSG("timer_thread terminated\n");
    return NULL;
}

/* start the timer thread */
int timer_start(void)
{
    if(!running) {
        int err;
        if(!pqueue_init())
            return 0; /* should not happen */
        err = pthread_create(&timer_thr, NULL, timer_thread, NULL);
        if(err != 0)
            perror("Error creating timer thread");
        else
            running = 1;
    }
    return running; /* if it's running the call succeed */
}

/* stop the timer thread */
int timer_stop(void)
{
    if(running) {
        int err;
        /* check if this is timer_thread (no synchronization needed) */
        int synchronous = pthread_equal(pthread_self(), timer_thr);
        terminate = 1;
        if(!synchronous) {
            pthread_mutex_lock(&tmr_mtx);
            pthread_cond_signal(&tmr_wakeup);
            pthread_mutex_unlock(&tmr_mtx);
            err = pthread_join(timer_thr, NULL);
            running = 0; /* timer_thread will stop sooner or later :) */
            if(err) {
                perror("Error joining timer thread");
                return 0;
            }
            pqueue_destroy();
        } else
            running = 0;
    }
    return 1;
}

/* schedule a job for execution at time tv */
int timer_schedule(struct timeval* tv, timer_handler th, void* param)
{
    int err;
    /* check if this is timer_thread (no synchronization needed) */
    int synchronous = pthread_equal(pthread_self(), timer_thr);
    job *jb = (job*) malloc(sizeof(job));
    if(!jb) return 0; /* should not happen */
    jb->tv = *tv;
    jb->th = th;
    jb->param = param;
    if(!synchronous) pthread_mutex_lock(&tmr_mtx);
    err = pqueue_push(jb);
    if(!synchronous) {
        pthread_cond_signal(&tmr_wakeup);
        pthread_mutex_unlock(&tmr_mtx);
    }
    if(!err)
        return -1; /* should not happen too */
    return jb->id;
}

int timer_cancel(int jobid)
{
    int* jb = (int*) jobs.v;
    job** pq = (job**) pqueue.v;
    int i;
    int synchronous;

    if(jobid >= jobs.size || jobid < 0 || (i=jb[jobid]) == -1)
        return -1;
    
    /* check if this is timer_thread (no synchronization needed) */
    synchronous = pthread_equal(pthread_self(), timer_thr);
    if(!synchronous) pthread_mutex_lock(&tmr_mtx);

    jb[jobid] = -1; /* mark as canceled */
    if(--pqueue.size > 0) {
        pq[i] = pq[pqueue.size];
        DBG(pq[pqueue.size] = NULL); /* offensive programming */
        if(i < pqueue.size)
            heap_down(pq, jb, i, pqueue.size);
    }
#ifndef NDEBUG /* offensive programming */
    else {
        assert(i == 0 && pqueue.size == 0);
        pq[0] = NULL;
    }
#endif

    if(!synchronous) {
        pthread_cond_signal(&tmr_wakeup);
        pthread_mutex_unlock(&tmr_mtx);
    }
    return 0;
}

int timer_changetime(int jobid, struct timeval* tv) /* N.B. O(log n) */
{
  int* jb = (int*) jobs.v;
  job** pq = (job**) pqueue.v;
  int i;
  int synchronous;

  if(jobid >= jobs.size || jobid < 0 || (i=jb[jobid]) == -1)
    return -1;
  
  /* check if this is timer_thread (no synchronization needed) */
  synchronous = pthread_equal(pthread_self(), timer_thr);
  if(!synchronous) pthread_mutex_lock(&tmr_mtx);

  pq[i]->tv = *tv;
  if(i > 0 && cmp_ent(pq[i], pq[i/2]) <= 0)
    heap_up(pq, jb, i);
  else
    heap_down(pq, jb, i, pqueue.size);

  if(!synchronous) {
      pthread_cond_signal(&tmr_wakeup);
      pthread_mutex_unlock(&tmr_mtx);
  }
  return 0;
}

int timer_changehndl(int jobid, timer_handler th) /* O(1) */
{
  int* jb = (int*) jobs.v;
  job** pq = (job**) pqueue.v;
  if(jobid >= jobs.size || jobid < 0 || jb[jobid] == -1)
    return -1;
  pq[jb[jobid]]->th = th;
  return 0;
}

int timer_changeparam(int jobid, void* param) /* O(1) */
{
  int* jb = (int*) jobs.v;
  job** pq = (job**) pqueue.v;
  if(jobid >= jobs.size || jobid < 0 || jb[jobid] == -1)
    return -1;
  pq[jb[jobid]]->param = param;
  return 0;
}

#ifndef NDEBUG

void show_heap_r(int i, int depth)
{
    pq_ent* pq = (pq_ent*)pqueue.v;
    if(i >= pqueue.size) return;
    printf("%*c%d: time = %.10gs, id = %d\n", depth, ' ',
       i, pq[i]->tv.tv_sec + 0.000001 * pq[i]->tv.tv_usec, pq[i]->id);
    show_heap_r(2 * i + 1, depth + 2);
    show_heap_r(2 * i + 2, depth + 2);
}

void show_heap(void) { show_heap_r(0, 0); }

#endif
